import 'package:flutter/painting.dart';

import '../extensions/extensions.dart';

import 'anchor.dart';
import 'cache_renderer_mixin.dart';
import 'renderer.dart';
import 'matrix_transformer.dart';

class Actor extends Renderer with CacheRendererMixin {
  final MatrixTransform transform;
  Anchor _anchor;
  Vector2 _size;

  Actor({
    double? angle,
    Vector2? position,
    Vector2? skew,
    Vector2? size,
    Anchor? anchor,
    Vector2? scale,
    double? priority,
  })  : _anchor = anchor ?? Anchor.center,
        _size = size ?? Vector2.zero(),
        transform = MatrixTransform(
          angle: angle,
          position: position,
          skew: skew,
          scale: scale,
        ),
        super(priority: priority) {
    transform.setOffsetBySizeAndAnchor(_size, _anchor);
  }

  void setFrom(Actor other) {
    anchor = other.anchor;
    size = other.size;

    angle = other.angle;
    scale = other.scale;
    skew = other.skew;
    position = other.position;
  }

  double get angle => transform.angle;
  set angle(double next) {
    if (transform.angle != next) {
      tick();
      onAngleChanged(transform.angle, transform.angle = next);
    }
  }

  void onAngleChanged(double prev, double next) {}

  Vector2 get position => transform.position;
  set position(Vector2 next) {
    if (transform.position != next) {
      tick();
      onPositionChanged(transform.position, transform.position = next);
    }
  }

  void onPositionChanged(Vector2 prev, Vector2 next) {}

  Vector2 get scale => transform.scale;
  set scale(Vector2 next) {
    if (transform.scale != next) {
      tick();
      onScaleChanged(transform.scale, transform.scale = next);
    }
  }

  void onScaleChanged(Vector2 prev, Vector2 next) {}

  Vector2 get skew => transform.skew;
  set skew(Vector2 next) {
    if (transform.skew != next) {
      tick();
      onSkewChanged(transform.skew, transform.skew = next);
    }
  }

  void onSkewChanged(Vector2 prev, Vector2 next) {}

  Vector2 get size => _size;
  set size(Vector2 next) {
    if (_size != next) {
      tick();
      transform.setOffsetBySizeAndAnchor(next, _anchor);
      onSizeChanged(_size, _size = next.clone());
    }
  }

  double get width => _size.x;
  double get height => _size.y;

  void onSizeChanged(Vector2 prev, Vector2 next) {}

  Anchor get anchor => _anchor;
  set anchor(Anchor next) {
    if (_anchor != next) {
      tick();
      transform.setOffsetBySizeAndAnchor(_size, next);
      onAnchorChanged(_anchor, _anchor = next);
    }
  }

  void onAnchorChanged(Anchor prev, Anchor next) {}

  @override
  void transformCanvas(Canvas canvas, VoidCallback callback) {
    canvas
      ..save()
      ..transform(transform.matrix.storage);
    callback();
    canvas.restore();
  }

  @override
  bool containsPoint(Vector2 point) {
    final local = absoluteToLocal(point);
    return (local.x >= 0) && (local.y >= 0) && (local.x < size.x) && (local.y < size.y);
  }

  Vector2 parentPositionOf(Vector2 point) => transform.localToGlobal(point);

  Vector2 parentPositionOfAnchor(Anchor anchor) {
    return anchor == this.anchor
        ? position
        : parentPositionOf(Vector2(anchor.x * size.x, anchor.y * size.y));
  }

  Vector2 absolutePositionOf(Vector2 point) {
    Vector2 parentPoint = parentPositionOf(point);
    Renderer? ancestor = parent;

    while (ancestor != null) {
      if (ancestor is Actor) {
        parentPoint = ancestor.parentPositionOf(parentPoint);
      }
      ancestor = ancestor.parent;
    }
    return parentPoint;
  }

  Vector2 absolutePositionOfAnchor(Anchor anchor) {
    return absolutePositionOf(Vector2(anchor.x * size.x, anchor.y * size.y));
  }

  Vector2 toLocal(Vector2 point) => transform.globalToLocal(point);

  Vector2 absoluteToLocal(Vector2 point) {
    Renderer? ancestor = parent;

    while (ancestor != null) {
      if (ancestor is Actor) {
        return toLocal(ancestor.absoluteToLocal(point));
      }
      ancestor = ancestor.parent;
    }
    return toLocal(point);
  }

  Rect toRect() => transform.toRect(_size);
  Rect toAbsoluteRect() => transform.toRect(_size, absolutePositionOf);

  double get absoluteAngle {
    return ancestors()
        .whereType<Actor>()
        .fold<double>(angle, (totalAngle, c) => totalAngle + c.angle);
  }

  Vector2 get absolutePosition => absolutePositionOfAnchor(anchor);
  Vector2 get center => parentPositionOfAnchor(Anchor.center);
  Vector2 get absoluteCenter => absolutePositionOfAnchor(Anchor.center);
}
