import 'package:flutter_test/flutter_test.dart' as flutter_test show test;
import 'package:flutter_test/flutter_test.dart' hide test;
import 'package:meta/meta.dart';

import '../theatre.dart';

extension TheatreFinds on CommonFinders {
  Finder byStage<T extends Stage>() {
    return find.byWidgetPredicate((widget) => widget is StageWidget<T>);
  }
}

typedef StageCreateFunction<T extends Stage> = T Function();
typedef VerifyFunction<T extends Stage> = dynamic Function(T);

typedef StageWidgetCreateFunction<T extends Stage> = StageWidget<T> Function(
  T stage,
);
typedef WidgetVerifyFunction<T extends Stage> = Future<void> Function(
  T,
  WidgetTester,
);
typedef WidgetSetupFunction<T extends Stage> = Future<void> Function(
  T,
  WidgetTester,
);
typedef PumpWidgetFunction<T extends Stage> = Future<void> Function(
  StageWidget<T>,
  WidgetTester tester,
);

class StageTester<T extends Stage> {
  final StageCreateFunction<T> createStage;

  final StageWidgetCreateFunction<T>? createStageWidget;

  final PumpWidgetFunction<T>? pumpWidget;

  final Vector2? canvasSize;

  StageTester(
    this.createStage, {
    this.canvasSize,
    this.createStageWidget,
    this.pumpWidget,
  });

  T initializeStage() {
    final stage = createStage();
    final size = canvasSize ?? Vector2.all(500);

    stage.onCanvasResize(size);
    stage.onMount();
    stage.update();

    return stage;
  }

  @isTest
  void test(String description, VerifyFunction<T> verify, {String? skip}) {
    flutter_test.test(
      description,
      () => verify(initializeStage()),
      skip: skip,
    );
  }

  @isTest
  void testStageWidget(
    String description, {
    WidgetSetupFunction<T>? setUp,
    WidgetVerifyFunction<T>? verify,
  }) {
    testWidgets(description, (tester) async {
      final stage = createStage();

      await tester.runAsync(() async {
        final stageWidget = createStageWidget?.call(stage) ?? StageWidget(stage: stage);

        final _pump = pumpWidget ??
            (StageWidget<T> _stageWidget, WidgetTester _tester) => _tester.pumpWidget(_stageWidget);

        await _pump(stageWidget, tester);
        await tester.pump();

        if (setUp != null) {
          await setUp.call(stage, tester);
          await tester.pump();
        }
      });

      if (verify != null) {
        await verify(stage, tester);
      }
    });
  }

  StageTester<T> configure({
    StageCreateFunction<T>? createStage,
    Vector2? canvasSize,
    StageWidgetCreateFunction<T>? createStageWidget,
    PumpWidgetFunction<T>? pumpWidget,
  }) {
    return StageTester<T>(
      createStage ?? this.createStage,
      canvasSize: canvasSize ?? this.canvasSize,
      createStageWidget: createStageWidget ?? this.createStageWidget,
      pumpWidget: pumpWidget ?? this.pumpWidget,
    );
  }
}
